import base64
import inspect
import io
import itertools
import logging
import os
import sys
import traceback
import typing as t
from functools import partial
from functools import reduce
from inspect import isfunction
from pathlib import Path

import boto3
from flask import Blueprint as FlaskBlueprint
from flask import Flask
from flask import current_app
from flask import json
from flask.blueprints import BlueprintSetupState
from flask.testing import FlaskClient
from werkzeug.datastructures import ImmutableDict
from werkzeug.datastructures import MultiDict
from werkzeug.datastructures import WWWAuthenticate
from werkzeug.exceptions import Forbidden
from werkzeug.exceptions import InternalServerError
from werkzeug.exceptions import Unauthorized
from werkzeug.routing import Rule

from .globals import request
from .utils import BIZ_BUCKET_HEADER_KEY
from .utils import BIZ_KEY_HEADER_KEY
from .utils import HTTP_METHODS
from .utils import create_cws_proxy
from .utils import get_app_stage
from .utils import get_cws_annotations
from .utils import is_arg_parameter
from .utils import is_kwarg_parameter
from .utils import make_absolute
from .utils import path_join
from .utils import trim_underscores
from .wrappers import CoworksMapAdapter
from .wrappers import CoworksRequest
from .wrappers import CoworksResponse
from .wrappers import TokenResponse

if t.TYPE_CHECKING:  # pragma: no cover
    from _typeshed.wsgi import StartResponse
    from _typeshed.wsgi import WSGIEnvironment
else:
    WSGIEnvironment = t.Any
    StartResponse = t.Any


#
# Decorators
#


def entry(fun: t.Callable | None = None, binary_headers: dict[str, str] | None = None,
          stage: str | t.Iterable[str] | None = None,
          no_auth: bool = False, no_cors: bool = True) -> t.Callable:
    """Decorator to create a microservice entry point from function name.

    :param fun: the entry function.
    :param binary_headers: force default content-type.
    :param stage: entry defined only for this stage(s).
    :param no_auth: set authorizer by default.
    :param no_cors: set CORS by default.
    """
    if fun is None:
        return partial(entry, binary_headers=binary_headers, stage=stage, no_auth=no_auth, no_cors=no_cors)

    def get_path(start):
        name_ = fun.__name__[start:]
        name_ = trim_underscores(name_)  # to allow several functions with d, Falseifferent args
        return name_.replace('_', '/')

    name = fun.__name__.upper()
    for method in HTTP_METHODS:
        if name == method:
            path = ''
            break
        if name.startswith(f'{method}_'):
            path = get_path(len(f'{method}_'))
            break
    else:
        method = 'POST'
        path = get_path(0)

    stage = [] if stage is None else stage

    fun.__CWS_METHOD = method  # type: ignore[attr-defined]
    fun.__CWS_PATH = path  # type: ignore[attr-defined]
    fun.__CWS_BINARY_HEADERS = binary_headers  # type: ignore[attr-defined]
    fun.__CWS_NO_AUTH = no_auth  # type: ignore[attr-defined]
    fun.__CWS_STAGES = stage if isinstance(stage, list) else [stage]  # type: ignore[attr-defined]
    fun.__CWS_NO_CORS = no_cors  # type: ignore[attr-defined]

    return fun


#
# Classes
#

class CoworksClient(FlaskClient):
    """Creates environment and complete request.
    """

    def __init__(self, *args: t.Any, aws_event=None, aws_context=None, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)

        headers: dict = aws_event['headers']
        request_context = aws_event['requestContext']

        method: str = request_context['httpMethod']
        scheme: str = headers.get('x-forwarded-proto', "https")
        host_name: str = headers['x-forwarded-host'] if 'x-forwarded-host' in headers else request_context['domainName']
        entry_path: str = request_context['entryPath']
        stage: str = request_context['stage']
        entry_path_parameters: dict = aws_event['entryPathParameters']
        query_string: dict = MultiDict(aws_event['multiValueQueryStringParameters'])

        is_encoded: bool = aws_event.get('isBase64Encoded', False)
        body = aws_event['body']
        if body and is_encoded:
            body = base64.b64decode(body)

        self.aws_environ = {
            "wsgi.input": io.BytesIO(b""),
            "wsgi.url_scheme": scheme,
            "REQUEST_SCHEME": scheme,
            "REQUEST_METHOD": method,
            "SERVER_NAME": host_name,
            "PATH_INFO": entry_path,

            'aws_event': aws_event,
            'aws_context': aws_context,
            'aws_stage': stage,
            "aws_entry_path": entry_path,
            "aws_entry_path_parameters": entry_path_parameters,
            'aws_query_string': query_string,
            'aws_body': body,
        }

    def open(self, *args, **kwargs) -> CoworksResponse:  # type: ignore[override]
        req = CoworksRequest(self.aws_environ, populate_request=False, shallow=True)
        return t.cast(CoworksResponse, super().open(req))

    def _copy_environ(self, other) -> dict:
        return {**other}


class Blueprint(FlaskBlueprint):
    """ Represents a blueprint, list of routes that will be added to microservice when registered.

    See :ref:`Blueprint <blueprint>` for more information.
    """

    def __init__(self, name: str | None = None, **kwargs):
        """Initialize a blueprint.

        :param kwargs: Other Flask blueprint parameters.
        """
        import_name = self.__class__.__name__.lower()
        super().__init__(name or import_name, import_name, **kwargs)

    def init_app(self, app):
        """".. deprecated:: 0.8.0"""
        ...

    def init_cli(self, app):
        """.. deprecated:: 0.8.0"""
        ...

    @property
    def logger(self) -> logging.Logger:
        return current_app.logger

    def make_setup_state(self, app: "TechMicroService",  # type: ignore[override]
                         options: dict, *args) -> BlueprintSetupState:
        """Stores creation state for deferred initialization."""
        state = super().make_setup_state(app, options, *args)

        # Defer blueprint route initialization.
        if not options.get('hide_routes', False):
            func = partial(TechMicroService.add_coworks_routes, state.app, state)
            app.deferred_init_routes_functions = itertools.chain(app.deferred_init_routes_functions, (func,))

        return state


class TechMicroService(Flask):
    """Simple tech microservice.

    See :ref:`tech` for more information.

    .. versionadded:: 0.8.0
       The `stage_prefixed` parameter was added.
    """

    # Maximume content length for writing response content in debug
    size_max_for_debug = 2000

    def __init__(self, name: str | None = None, stage_prefixed: bool = True, **kwargs) -> None:
        """ Initialize a technical microservice.
        :param name: Name used to identify the microservice.
        :param stage_prefixed: if accessed with stage or not.
        :param kwargs: Other Flask parameters.
        """
        self.default_config = ImmutableDict({
            **self.default_config,
        })

        name = name or self.__class__.__name__.lower()
        super().__init__(import_name=name, static_folder=None, **kwargs)

        self.request_class: type[CoworksRequest] = CoworksRequest
        self.response_class: type[CoworksResponse] = CoworksResponse

        self.deferred_init_routes_functions: t.Iterable[t.Callable] = []
        self._cws_app_initialized = False
        self._cws_conf_updated = False
        self.__aws_url_map: dict | None = None
        self.__stage_prefixed = stage_prefixed

        @self.before_request
        def before():
            self._check_token()

    def init_app(self):
        """Called to finalize the application initialization.
        Mainly to get external variables defined specifically for a workspace.

        .. deprecated:: 0.8.0
        """

    def init_cli(self):
        """Called only on cli command.
        Mainly externalized to allow specific import not needed on deployed implementation.

        .. deprecated:: 0.8.0
        """

    def app_context(self):
        """Override to initialize coworks microservice.
        """
        self._init_app(False)
        return super().app_context()

    def create_url_adapter(self, _request: CoworksRequest | None):  # type: ignore[override]
        if _request and _request.aws_event:
            self.subdomain_matching = True
            return CoworksMapAdapter(_request.environ, self.url_map, self.aws_url_map, self.__stage_prefixed)
        return super().create_url_adapter(_request)

    def cws_client(self, aws_event, aws_context):
        """CoWorks client used by the lambda call.
        """
        return CoworksClient(self, CoworksResponse, use_cookies=True, aws_event=aws_event, aws_context=aws_context)

    @property
    def header_autorizer_token(self):
        """The identity source for which authorization is requested usefull only on local.
        Usually defined in API GAateway Authorizer resource."""
        return request.headers.get('Authorization')

    @property
    def in_lambda_context(self):
        """Defined as a property to be read only."""
        return self._in_lambda_context

    @property
    def aws_url_map(self) -> dict[str, list[Rule]]:
        if self.__aws_url_map is None:
            self.__aws_url_map = {}
            for rule in self.url_map.iter_rules():
                entry_path = rule.rule.replace('<', '{').replace('>', '}')
                if entry_path in self.__aws_url_map:
                    self.__aws_url_map[entry_path].append(rule)
                else:
                    self.__aws_url_map[entry_path] = [rule]
        return self.__aws_url_map

    @property
    def routes(self) -> list[str]:
        """Returns the list of routes defined in the microservice.
        """
        return [k for k in self.aws_url_map]

    def token_authorizer(self, token: str) -> bool | str:
        """Defined the authorization process.

        If the returned value is False, all routes for all stages are denied.
        If the returned value is True, all routes for all stages are accepted.
        If the returned value is a string, then it must be a stage name and all routes are accepted for this stage.

        By default, just using the token authentification process (defined from terraform template).
        """

        if token is None:
            raise Unauthorized(www_authenticate=WWWAuthenticate(auth_type="basic"))

        return token == os.getenv('TOKEN')

    def auto_find_instance_path(self):
        """Instance path may be redefined by an environment variable.
        """
        instance_relative_path = os.getenv('INSTANCE_RELATIVE_PATH', '')
        path = Path(self.root_path) / instance_relative_path
        return path.as_posix()

    def store_response(self, resp: t.Union[dict, bytes], request_headers):
        """Store microservice response in S3 for biz task sequence.

        May be redefined for another storage in asynchronous call.
        """

        bucket = key = ''
        try:
            bucket = request_headers.get(self.config['X-CWS-S3Bucket'].lower())
            key = request_headers.get(self.config['X-CWS-S3Key'].lower())
            if bucket and key:
                content = json.dumps(resp).encode() if isinstance(resp, dict) else resp
                buffer = io.BytesIO(content)
                buffer.seek(0)
                aws_s3_session = boto3.session.Session()
                aws_s3_session.client('s3').upload_fileobj(buffer, bucket, key)
                self.logger.debug(f"Response stored in {bucket}/{key}")
            else:
                self.logger.error("Cannot store response as no bucket or key")
        except Exception as e:
            self.logger.error(f"Exception when storing response in {bucket}/{key} : {str(e)}")

    def _init_app(self, in_lambda):
        """Finalize the app initialization.
        """
        if not self._cws_app_initialized:
            self._in_lambda_context = in_lambda
            self._update_config(in_lambda=in_lambda)
            self.add_coworks_routes()
            for fun in self.deferred_init_routes_functions:
                fun()

            if os.getenv("FLASK_RUN_FROM_CLI"):
                self.init_cli()

            self._cws_app_initialized = True

    def __call__(self, arg1, arg2) -> dict:
        """Main microservice entry point.
        """
        in_flask = isfunction(arg2)
        self._init_app(not in_flask)

        # Lambda event call or Flask call
        if in_flask:
            res = self._flask_handler(arg1, arg2)
        else:
            res = self._lambda_handler(arg1, arg2)

        return res

    def _lambda_handler(self, event: dict[str, t.Any], context: dict[str, t.Any]):
        """Lambda handler.
        """
        if event.get('type') == 'TOKEN':
            return self._token_handler(event, context)
        return self._api_handler(event, context)

    def _token_handler(self, aws_event: dict[str, t.Any], aws_context: dict[str, t.Any]) -> dict:
        """Authorization token handler.
        """
        self.logger.warning(f"Calling {self.name} for authorization : {aws_event}")

        try:
            res = self.token_authorizer(aws_event['authorizationToken'])
            self.logger.warning(f"Token authorizer return is : {res}")
            return TokenResponse(res, aws_event['methodArn']).json
        except Exception as e:
            self.full_logger_error(e)
            return TokenResponse(False, aws_event['methodArn']).json

    def _api_handler(
            self, aws_event: dict[str, t.Any], aws_context: dict[str, t.Any]
    ) -> t.Optional[t.Union[dict, str]]:
        """API handler.
        """
        self.logger.warning(f"Calling {self.name} by api : {aws_event}")

        # Transforms as simple client call and manage exception if needed
        try:
            with self.cws_client(aws_event, aws_context) as c:

                # Get Flask return as dict or binary content
                resp = self._convert_to_lambda_response(c.open())

                # Strores response in S3 if asynchronous call
                invocation_type = aws_event['headers'].get('invocationtype')
                if invocation_type == 'Event':
                    self.store_response(resp, aws_event['headers'])

                # Encodes binary content
                if not isinstance(resp, dict):
                    str_resp = base64.b64encode(resp).decode('ascii')
                    content_length = len(str_resp) if self.logger.getEffectiveLevel() == logging.DEBUG else "N/A"
                    self.logger.warning(f"API returns binary content [length: {content_length}]")
                    return str_resp

                # Adds trace
                self.logger.warning(f"API returns code {resp.get('statusCode')} and headers {resp.get('headers')}")
                self.logger.warning(f"API body: {str(resp.get('body'))[:self.size_max_for_debug]}")

                return resp

        except Exception as e:
            # no more in app context so must print for trace
            print(f"Exception in {self.name} : {e}")
            parts = ["Traceback (most recent call last):\n"]
            parts.extend(traceback.format_stack(limit=15)[:-2])
            parts.extend(traceback.format_exception(*sys.exc_info())[1:])
            trace = reduce(lambda x, y: x + y, parts, "")
            print(f"Traceback: {trace}")
            headers = {'content_type': "application/json"}
            return self._aws_payload(str(e), InternalServerError.code, headers)

    def _flask_handler(self, environ: WSGIEnvironment, start_response: StartResponse):
        """Flask handler.
        """
        return self.wsgi_app(environ, start_response)

    def _update_config(self, *, in_lambda: bool):
        if not self._cws_conf_updated:
            # Set predefined environment variables
            self.config['X-CWS-S3Bucket'] = BIZ_BUCKET_HEADER_KEY
            self.config['X-CWS-S3Key'] = BIZ_KEY_HEADER_KEY

            self._cws_conf_updated = True

    def _check_token(self):
        """Simulates the authorization process of lambda if not in lambda context."""
        if not self.in_lambda_context:

            # Get no_auth option for this entry
            no_auth = False
            if request.url_rule:
                view_function = self.view_functions.get(request.url_rule.endpoint, None)
                if view_function:
                    no_auth = get_cws_annotations(view_function, '__CWS_NO_AUTH', False)

            # Checks token if authorization needed
            if not no_auth:
                try:
                    token = self.header_autorizer_token
                    valid = self.token_authorizer(token)
                    if not valid:
                        raise Forbidden()
                except Exception as e:
                    current_app.logger.exception(e)
                    raise Forbidden()

    def _convert_to_lambda_response(self, resp: CoworksResponse) -> t.Union[dict, bytes]:
        """Convert Lambda response (dict for JSON or text result, bytes for binary content)."""

        # returns JSON structure
        if resp.is_json:
            try:
                return self._aws_payload(resp.json, resp.status_code, resp.headers)
            except (Exception,):
                resp.mimetype = "text/plain"

        # returns simple string JSON structure
        if resp.mimetype and resp.mimetype.startswith('text'):
            try:
                return self._aws_payload(resp.get_data(True), resp.status_code, resp.headers)
            except ValueError:
                pass

        # returns direct payload
        data = resp.get_data()
        if not isinstance(data, bytes):
            msg = f'Expected bytes type for body with binary Content-Type. Got {type(data)} type body instead.'
            raise ValueError(msg)
        return data

    def _aws_payload(self, body, status_code, headers):
        return {
            "statusCode": status_code,
            "headers": {k: v for k, v in headers.items()},
            "body": body,
            "isBase64Encoded": False,
        }

    def full_logger_error(self, error):
        self.logger.error(f"Exception in {self.name} : {error}")
        parts = ["Traceback (most recent call last):\n"]
        parts.extend(traceback.format_stack(limit=15)[:-2])
        parts.extend(traceback.format_exception(*sys.exc_info())[1:])
        trace = reduce(lambda x, y: x + y, parts, "")
        self.logger.error(f"Traceback: {trace}")

    def add_coworks_routes(self, bp_state: BlueprintSetupState | None = None) -> None:
        """ Creates all routes for a microservice.

        :param bp_state: The blueprint state.
        """

        # Adds entrypoints
        stage = get_app_stage()
        scaffold = bp_state.blueprint if bp_state else self
        method_members = inspect.getmembers(scaffold.__class__, lambda x: inspect.isfunction(x))
        methods = [fun for _, fun in method_members if get_cws_annotations(fun, '__CWS_METHOD')]
        for fun in methods:

            # the entry is not defined for this stage
            stages = get_cws_annotations(fun, '__CWS_STAGES')
            if stages and stage not in stages:
                continue

            method = get_cws_annotations(fun, '__CWS_METHOD')
            entry_path = path_join(get_cws_annotations(fun, '__CWS_PATH'))

            # Get parameters
            sig = inspect.signature(fun)
            param_names = list(sig.parameters.keys())[1:]  # skip self parameter
            generic_kwargs = None
            if len(param_names) > 0:
                last_param_name = param_names[-1]
                last_param = sig.parameters[param_names[-1]]
                if last_param.kind == last_param.VAR_KEYWORD:
                    generic_kwargs = last_param_name
                    param_names = param_names[:-1]
            args = [n for n in param_names if is_arg_parameter(sig.parameters[n])]
            for index, arg in enumerate(args):
                entry_path = path_join(entry_path, f"/<{arg}>")
            kwargs = {n: sig.parameters[n] for n in param_names if is_kwarg_parameter(sig.parameters[n])}

            proxy = create_cws_proxy(scaffold, fun, args, kwargs, generic_kwargs)
            proxy.__CWS_BINARY_HEADERS = get_cws_annotations(fun, '__CWS_BINARY_HEADERS')
            proxy.__CWS_NO_AUTH = get_cws_annotations(fun, '__CWS_NO_AUTH')
            proxy.__CWS_NO_CORS = get_cws_annotations(fun, '__CWS_NO_CORS')
            fun.__CWS_FROM_BLUEPRINT = bp_state.blueprint.name if bp_state else None

            prefix = f"{bp_state.blueprint.name}." if bp_state else ''
            endpoint = f"{prefix}{fun.__name__}"

            # Creates the entry
            url_prefix = bp_state.url_prefix if bp_state else None
            rule = make_absolute(entry_path, url_prefix)
            for r in self.url_map.iter_rules():
                if r.rule == rule and method in (r.methods or []):
                    raise AssertionError(f"Duplicate route {rule}")
            try:
                self.add_url_rule(rule=rule, view_func=proxy, methods=[method], endpoint=endpoint, strict_slashes=False)
            except AssertionError:
                raise



import inspect
import os
import sys
import typing as t
from collections import defaultdict
from inspect import Parameter
from textwrap import dedent

import markdown
from flask import abort
from flask import current_app
from flask import render_template_string
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape
from jsonapi_pydantic.v1_0 import Resource
from jsonapi_pydantic.v1_0 import TopLevel
from pydantic import BaseModel
from werkzeug.exceptions import NotFound

from coworks import Blueprint
from coworks import entry
from coworks.utils import get_cws_annotations


class Admin(Blueprint):
    """Administration blueprint to get details on the microservice containing it.
    """

    def __init__(self, name: str = 'admin', models: t.Optional[list[type[BaseModel]]] = None, **kwargs):
        super().__init__(name=name, **kwargs)
        self.models = {model.__name__: model for model in models} if models else {}

    @entry(no_auth=True, no_cors=True)
    def get(self):
        """Returns the markdown documentation associated to this microservice.
        """
        md = getattr(current_app, 'doc_md', None)
        if not md:
            md = getattr(current_app.__class__, 'DOC_MD', None)
        if not md and current_app.__class__.__doc__:
            md = current_app.__class__.__doc__.replace('\n', ' ').strip()
        content = markdown.markdown(md, extensions=['fenced_code']) if md else ""

        routes = dict(sorted(self.get_route(blueprint="__all__").items()))
        bottom = render_template_string(self.routes_template, routes=routes)

        headers = {
            "Content-Type": 'text/html; charset=utf-8'
        }
        content = self.header_template + '<hr/>' + content + '<hr/>' + bottom + '\n'
        return content, 200, headers

    @entry(no_auth=True, no_cors=True)
    # @jsonapi(type)
    def get__schema(self, model):
        """Returns the JSON schemas of the model.

        :param model: Model's name.
       """
        if model in self.models:
            schema = self.models[model].schema()
            return TopLevel(
                data=Resource(id=model, type="JsonApiSchema", attributes={"schema": schema})).model_dump_json()
        abort(404)

    @entry(stage="dev")
    def get_route(self, prefix=None, blueprint=None):
        """Returns the list of entrypoints with signature.

        :param prefix: Prefix path to limit the number of returned routes.
        :param blueprint: Show named blueprint routes if defined ('__all__' or blueprint name).
        """
        if blueprint and (blueprint != '__all__' and blueprint not in current_app.blueprints):
            raise NotFound(f"Undefined blueprint {blueprint}")

        routes: dict[str, dict[str, str]] = defaultdict(dict)

        for rule in current_app.url_map.iter_rules():

            # Must return only prefixed routes
            if prefix:
                if rule.rule.startswith(prefix):
                    self.add_route_from_rule(routes, rule)
                else:
                    continue

            function_called = current_app.view_functions[rule.endpoint]
            from_blueprint = get_cws_annotations(function_called, '__CWS_FROM_BLUEPRINT')

            # Must return only blueprint routes
            if blueprint:
                if blueprint == '__all__' or from_blueprint == blueprint:
                    self.add_route_from_rule(routes, rule)
                else:
                    continue

            if from_blueprint is None:
                self.add_route_from_rule(routes, rule)

        return routes

    def get_proxy(self):
        """Returns the calling context."""
        env = Environment(
            loader=PackageLoader(sys.modules[__name__].__name__),
            autoescape=select_autoescape(['html', 'xml']))
        env.filters["signature"] = inspect.signature
        env.filters["positional_params"] = positional_params
        env.filters["keyword_params"] = keyword_params

        data = {
            'name': current_app.name,
            'entries': current_app.url_map,
        }
        template = env.get_template("proxy.j2")
        return template.render(**data)

    def add_route_from_rule(self, routes, rule):
        route = {}
        for http_method in rule.methods:
            if http_method not in ['HEAD', 'OPTIONS']:
                function_called = current_app.view_functions[rule.endpoint]
                route[http_method] = {
                    'signature': get_signature(function_called),
                    'endpoint': rule.endpoint,
                    'binary_headers': get_cws_annotations(function_called, '__CWS_BINARY_HEADERS'),
                    'no_auth': get_cws_annotations(function_called, '__CWS_NO_AUTH'),
                    'no_cors': get_cws_annotations(function_called, '__CWS_NO_CORS'),
                }

                from_blueprint = get_cws_annotations(function_called, '__CWS_FROM_BLUEPRINT', None)
                if from_blueprint:
                    route[http_method]['blueprint'] = from_blueprint

                doc = inspect.getdoc(function_called)
                if doc:
                    docstring = doc.replace('\n', ' ').split(':param ')
                    route[http_method]['doc'] = docstring[0].strip()
                    if len(docstring) > 1:
                        # noinspection PyTypeChecker
                        route[http_method]['params'] = docstring[1:]

        routes[rule.rule].update(route)

    @property
    def header_template(self) -> str:
        deployed = os.getenv("CWS_DATETIME", None)
        description = current_app.name
        if deployed:
            lambda_name = os.getenv("CWS_LAMBDA", description)
            description = f"{lambda_name} deployed {deployed}"
        return dedent(f"""<div style=\"display:flex;justify-content:space-between;\">
            <span style=\"font-size:xx-large;font-weight:bold\">{current_app.__class__.__name__}</span>
            <img style=\"margin-bottom:auto;width:100px;\"
            src=\"https://github.com/gdoumenc/coworks/raw/dev/docs/img/coworks.png\"/>
            </div><div style=\"display:flex;flex-direction:row-reverse;font-size:small;margin-top:5px;\">
            {description}</div>""")

    @property
    def routes_template(self) -> str:
        return dedent(
            """<style type="text/css">ul.nobull {list-style-type: none;}</style>
            <ul class="nobull">{% for entry,route in routes.items() %}
                <li>{{ entry }} : <ul>{% for method,info in route.items() %}
                    <li><i>{{ method }}{{ info.signature }}[endpoint: {{info.endpoint}}]</i> : {{ info.doc }}
                    <ul class="nobull">{% for param in info.params %}<li><i>{{ param }}</i>{% endfor %}</ul>
                {% endfor %}</li></ul></li>
            {% endfor %}</ul>"""
        )


def get_signature(func: t.Callable) -> str:
    sig = ""
    params = inspect.signature(func).parameters
    for i, (k, p) in enumerate(params.items()):
        if i == 0:
            continue
        sp = k
        if p.annotation != Parameter.empty:
            sp = f"{sp}:{str(p.annotation)}"
        if p.default != Parameter.empty:
            sp = f"{sp}={p.default}"
        sig = f"{sp}" if i == 1 else f"{sig}, {sp}"
    return f"({sig})"


def positional_params(func: t.Callable) -> str:
    res = ''
    params = inspect.signature(func).parameters
    for i, (k, p) in enumerate(params.items()):
        if i == 0 or p.kind not in [Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD,
                                    Parameter.VAR_POSITIONAL]:
            continue
        if p.kind == Parameter.POSITIONAL_OR_KEYWORD and p.default != Parameter.empty:
            continue
        sp = f"'{k}' : {k}"
        res = f"{sp}" if i == 1 else f"{res}, {sp}"
    return res


def keyword_params(func: t.Callable) -> str:
    res = ''
    params = inspect.signature(func).parameters
    for i, (k, p) in enumerate(params.items()):
        if i == 0 or p.kind not in [Parameter.KEYWORD_ONLY, Parameter.POSITIONAL_OR_KEYWORD, Parameter.VAR_KEYWORD]:
            continue
        if p.kind == Parameter.POSITIONAL_OR_KEYWORD and p.default == Parameter.empty:
            continue
        sp = f"'{k}' : {k}"
        res = f"{sp}" if i == 1 else f"{res}, {sp}"
    return res



import multiprocessing
import os
import time
from unittest import mock

import requests

from coworks.cws.client import CwsScriptInfo


class TestClass:

    @mock.patch.dict(os.environ, {"FLASK_RUN_FROM_CLI": "false"})
    def test_run_simple(self, samples_docs_dir, unused_tcp_port, monkeypatch):
        info = CwsScriptInfo(project_dir='tech')
        info.app_import_path = "hello:app"
        app = info.load_app()
        server = multiprocessing.Process(target=run_server, args=(app, unused_tcp_port), daemon=True)
        server.start()
        counter = 1
        time.sleep(counter)
        while not server.is_alive() and counter < 3:
            time.sleep(counter)
            counter += 1
        response = requests.get(f'http://localhost:{unused_tcp_port}/', headers={'Authorization': "token"})
        assert response.text == "Hello world.\n"
        server.terminate()


def run_server(app, port):
    print(f"Server starting on port {port}")
    app.run(host='localhost', port=port, use_reloader=False, debug=False)



import os

from aws_xray_sdk.core import xray_recorder
from flask import request
from okta.client import Client
from okta.okta_object import OktaObject
from werkzeug.exceptions import InternalServerError

from coworks import Blueprint
from coworks import entry
from coworks.extension.xray import XRay


class OktaClient(Client):
    """Okta client extended to allow next call event on new client defined."""

    @XRay.capture(xray_recorder)
    async def next(self, next):
        req, error = await self._request_executor.create_request("GET", next, {}, {})
        if error:
            return None, error
        response, error = await self._request_executor.execute(req)
        try:
            result = []
            for item in response.get_body():
                result.append(OktaDict(self.form_response_body(item)))
        except Exception as error:
            return None, error
        return result, response, None


class OktaResponse:
    """Class to manipulate results from okta client.
    value: value returned as a list of dict.
    err: Not None if error.
    next: Next url to be called if wanted more results.

    OktaRespons are used as global variables from asynchronous functions.

    To set the value from an asynchronous function:
    resp.set(await self.okta_client.function(query_parameters))

    The property response must be used as microservice returned value:
    return resp.response

    To combine results in microservice:
    return OktaResponse.combine({'user': resp_user, 'groups': resp_groups})
    """

    def __init__(self):
        self.api_resp = self.next_url = self.error = None

    @XRay.capture(xray_recorder)
    def set(self, await_result, fields=None):
        """Set the values from the result. Keep only specific fieds if defined."""

        if len(await_result) == 3:
            _, self.api_resp, self.error = await_result
        else:
            self.api_resp, self.error = await_result

        # noinspection PyProtectedMember
        self.next_url = self.api_resp._next if self.api_resp else None

    @property
    def body(self):
        """Get OKTA body response."""
        return self.api_resp.get_body() if self.api_resp else None

    @property
    def response(self):
        """Cast the Okta response as microservice response."""
        if self.error:
            return self.error.message, self.error.status
        return {'value': self.body, 'next': self.next_url}


class Okta(Blueprint):

    def __init__(self, name: str = 'okta',
                 env_url_var_name: str = '', env_token_var_name: str = '',
                 env_var_prefix: str = "OKTA", **kwargs):
        super().__init__(name=name, **kwargs)
        if env_var_prefix:
            self.env_url_var_name = f"{env_var_prefix}_URL"
            self.env_token_var_name = f"{env_var_prefix}_TOKEN"
        else:
            self.env_url_var_name = env_url_var_name
            self.env_token_var_name = env_token_var_name

        self.org_url = os.getenv(self.env_url_var_name)
        if not self.org_url:
            raise InternalServerError(f"Environment var {self.env_url_var_name} undefined.")
        config = {
            'orgUrl': self.org_url,
            'token': os.getenv(self.env_token_var_name)
        }
        self.okta_client = OktaClient(config)

    @entry
    def get_event_verify(self):
        """Entry for Okta webhook verification."""
        test_value = request.headers.get('x-okta-verification-challenge')
        return {"verification": test_value}


class OktaDict(OktaObject):
    """Simplified generic okta object."""

    def __init__(self, values):
        super().__init__()
        self.values = values

    def as_dict(self):
        return {k: v for k, v in self.values.items() if k != "links"}



from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.sensors.base import BaseSensorOperator
from airflow.sensors.base import poke_mode_only
from airflow.utils.decorators import apply_defaults

from coworks.biz.operators import TechMicroServiceOperator


@apply_defaults
@poke_mode_only
class TechMicroServiceSensor(BaseSensorOperator, TechMicroServiceOperator):
    """ Sensor to call a TechMicroservice until call succeed.

    Same parameters as TechMicroServiceOperator except 'asynchronous' and 'raise_400_errors'
    """

    def __init__(self, poke_interval: float = 30, **kwargs):
        super().__init__(asynchronous=False, raise_400_errors=False, poke_interval=poke_interval, **kwargs)

    def poke(self, context):
        """Set microservice context and execute it eache time."""
        self.pre_execute(context)
        res = self._call_cws(context)
        return res.ok


@apply_defaults
class AsyncTechMicroServiceSensor(BaseSensorOperator):
    """ Sensor to wait until an asynchronous TechMicroservice call is ended.

    :param cws_task_id: the tech microservice task_id awaited.
    :param aws_conn_id: AWS S3 connection.
    """

    def __init__(self, *, cws_task_id: str, aws_conn_id: str = 'aws_s3', **kwargs):
        super().__init__(**kwargs)
        self.cws_task_id = cws_task_id
        self.aws_conn_id = aws_conn_id

    def poke(self, context):
        ti = context['ti']
        bucket_name = ti.xcom_pull(task_ids=self.cws_task_id, key='bucket')
        bucket_key = ti.xcom_pull(task_ids=self.cws_task_id, key='key')

        self.log.info('Poking for key : s3://%s/%s', bucket_name, bucket_key)
        hook = S3Hook(aws_conn_id=self.aws_conn_id)
        return hook.check_for_key(bucket_key, bucket_name)



import os
import smtplib
from email import message
from io import BytesIO
from unittest import mock

import pytest

from coworks import TechMicroService
from coworks.blueprint.mail_blueprint import Mail

smtp_mock = mock.MagicMock()
smtp_mock.return_value.__enter__.return_value.login = login_mock = mock.Mock()
smtp_mock.return_value.__enter__.return_value.send_message = send_mock = mock.Mock()

email_mock = mock.MagicMock()
email_mock.return_value.add_attachment = add_mock = mock.Mock()


class MailMS(TechMicroService):

    def __init__(self, **mail_names):
        super().__init__('mail')
        self.register_blueprint(Mail(**mail_names))

    def _check_token(self):
        return True


class TestClass:

    def test_wrong_init(self):
        with pytest.raises(RuntimeError) as pytest_wrapped_e:
            app = MailMS()
            with app.test_client() as c:
                response = c.post('/send')
        assert pytest_wrapped_e.type == RuntimeError

    @mock.patch.dict(os.environ, {
        "SMTP_SERVER": "mail.test.com:587",
        "SMTP_LOGIN": "myself@test.com",
        "SMTP_PASSWD": "passwd"
    })
    def test_wrong_params(self):
        mail_names = {
            'env_server_var_name': 'SMTP_SERVER',
            'env_login_var_name': 'SMTP_LOGIN',
            'env_passwd_var_name': 'SMTP_PASSWD',
        }
        app = MailMS(**mail_names)
        with app.test_client() as c:
            data = {
                'subject': "Test",
            }
            response = c.post('/send', data=data)
            assert response.status_code == 400
            assert response.get_data(as_text=True) == "From address not defined (from_addr:str)"

    @mock.patch.dict(os.environ, {
        "SMTP_SERVER": "mail.test.com:587",
        "SMTP_LOGIN": "myself@test.com",
        "SMTP_PASSWD": "passwd"
    })
    def test_wrong_params(self, auth_headers):
        app = MailMS(env_var_prefix='SMTP')
        with app.test_client() as c:
            data = {
                'subject': "Test",
            }
            response = c.post('/send', data=data, headers=auth_headers)
            assert response.status_code == 400
            assert "From address not defined (from_addr:str)" in response.get_data(as_text=True)

    @mock.patch.dict(os.environ, {
        "SMTP_SERVER": "mail.test.com:587",
        "SMTP_LOGIN": "myself@test.com",
        "SMTP_PASSWD": "passwd"
    })
    @mock.patch.object(smtplib, 'SMTP', smtp_mock)
    def test_send_text(self, auth_headers):
        app = MailMS(env_var_prefix='SMTP')
        with app.test_client() as c:
            data = {
                'subject': "Test",
                'from_addr': "from@test.fr",
                'to_addrs': "to@test.fr",
            }
            response = c.post('/send', data=data, headers=auth_headers)
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "Mail sent to to@test.fr"
            login_mock.assert_called_with('myself@test.com', 'passwd')
            send_mock.assert_called_once()

    @mock.patch.dict(os.environ, {
        "SMTP_SERVER": "mail.test.com:587",
        "SMTP_LOGIN": "myself@test.com",
        "SMTP_PASSWD": "passwd"
    })
    @mock.patch.object(smtplib, 'SMTP', smtp_mock)
    @mock.patch.object(message, 'EmailMessage', email_mock)
    def test_send_attachment(self, auth_headers):
        app = MailMS(env_var_prefix='SMTP')
        with app.test_client() as c:
            file = BytesIO(b"hello {{ world_name }}")
            data = {
                'subject': "Test",
                'from_addr': "from@test.fr",
                'to_addrs': "to@test.fr",
                'attachments': [(file, 'file.txt')]
            }
            response = c.post('/send', data=data, headers=auth_headers)
            assert response.status_code == 200
            login_mock.assert_called_with('myself@test.com', 'passwd')
            add_mock.assert_called_once()



import os
from unittest import mock

from flask import url_for

from ..blueprint.blueprint import BP, InitBP
from ..ms import SimpleMS


@mock.patch.dict(os.environ, {"TOKEN": "token"})
class TestClass:

    def test_request(self):
        app = SimpleMS()
        app.register_blueprint(BP())
        with app.test_client() as c:
            response = c.get('/', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'get'
            response = c.get('/test/3', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "blueprint BP 3"
            response = c.get('/extended/test/3', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'blueprint extended test 3'

    def test_prefix(self):
        app = SimpleMS()
        app.register_blueprint(BP(), url_prefix="/prefix")
        with app.test_client() as c:
            response = c.get('/prefix/test/3', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "blueprint BP 3"
            response = c.get('/prefix/extended/test/3', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'blueprint extended test 3'

    def test_url_for(self):
        app = SimpleMS()
        app.register_blueprint(BP(), url_prefix="/prefix")
        with app.test_client() as c:
            response = c.get('/prefix/test/3', headers={'Authorization': 'token'})
            assert url_for('get') == '/'
            assert url_for('bp.get_test', index=2) == '/prefix/test/2'

    def test_before_activation(self):
        app = SimpleMS()
        init_bp = InitBP()
        app.register_blueprint(init_bp, url_prefix="/prefix")
        with app.test_client() as c:
            response = c.get('/prefix/test/3', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "blueprint BP 3"
            init_bp.do_before_activation.assert_called_once()
            init_bp.do_after_activation.assert_called_once()



import io
import os
from unittest import mock

from coworks import TechMicroService
from coworks import entry
from coworks.coworks import CoworksResponse
from ..ms import SimpleMS


class ParamMS(TechMicroService):
    value = "123"

    @entry
    def get(self, str):
        return str

    @entry
    def get_concat(self, str1, str2):
        return str1 + str2

    @entry
    def get_value(self):
        return self.value

    @entry
    def put_value(self, value=None):
        self.value = value
        return self.value

    @entry
    def get_param(self, str1, param1='default1', param2='default2'):
        return str1 + str(param1) + param2

    @entry
    def post_params(self, **kwargs: dict):
        return {
            'keys': [k for k in kwargs.keys()],
        }


class TupleReturnedMS(TechMicroService):

    @entry
    def get(self):
        return 'ok', 200

    @entry
    def get_json(self):
        return {'value': 'ok'}, 200

    @entry
    def get_resp(self, str):
        return CoworksResponse(str, 200)

    @entry
    def get_error(self, str):
        return str, 300

    @entry
    def get_tuple(self, str):
        return str, 200, {'x-test': 'true'}


class AmbiguousMS(TechMicroService):
    @entry
    def get(self, uid):
        return uid, 200

    @entry
    def get_test(self):
        return {'value': 'ok'}, 200


@mock.patch.dict(os.environ, {"TOKEN": "token"})
class TestClass:

    def test_request_arg(self):
        app = SimpleMS()
        with app.test_client() as c:
            response = c.get('/', headers={'Accept': 'text/plain', 'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get"
            assert 'Content-Type' in response.headers
            assert response.headers['Content-Type'] == 'application/json'
            assert 'Content-Length' in response.headers
            assert response.headers['Content-Length'] == str(len(response.get_data(as_text=True)))
            response = c.post('/', headers={'Authorization': 'token'})
            assert response.status_code == 405
            response = c.get('/get1', headers={'Authorization': 'token'})
            assert response.status_code == 404
            response = c.get('/content', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get content"
            response = c.get('/content/3', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get content with 3"
            response = c.get('/content/3/other', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get content with 3 and other"
            response = c.post('/content', json={"other": 'other'}, headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "post content without value but other"
            response = c.post('/content/3', json={"other": 'other'}, headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "post content with 3 and other"
            response = c.post('/content/3', json="other", headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "post content with 3 and other"
            response = c.post('/content/3', json={"other": 'other', "value": 5}, headers={'Authorization': 'token'})
            assert response.status_code == 422

    def test_request_kwargs(self):
        app = SimpleMS()
        with app.test_client() as c:
            response = c.get('/kwparam1?value=5', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with only 5"
            response = c.get('/kwparam1?other=other&value=5', headers={'Authorization': 'token'})
            assert response.status_code == 422
            response = c.get('/kwparam1?value=5', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with only 5"
            response = c.get('/kwparam1', json={"other": 'other', "value": 5}, headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with only 0"

            response = c.get('/kwparam2?other=other', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with 0 and ['other']"
            response = c.get('/kwparam2?other=other&value=5', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with 5 and ['other']"
            response = c.get('/kwparam2', json={"other": 'other', "value": 5}, headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with 0 and []"
            response = c.put('/kwparam2', json={"other": 'other', "value": 5}, headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with 5 and ['other']"
            response = c.put('/kwparam2?other=other&value=5', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with 0 and []"

            response = c.post('/kwparam2/5', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with 5 and []"
            response = c.post('/kwparam2/5', json={"other": 'other'}, headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with 5 and ['other']"
            response = c.post('/kwparam2/5?other=other', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "get **param with 5 and []"

            response = c.get('/extended/content', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "hello world"

    def test_request_form_data(self):
        """normal API call."""
        app = ParamMS()
        with app.test_client() as c:
            files = {
                'template': io.BytesIO(b"hello {{ world_name }}"),
            }
            data = {
                'key': 'value',
                'template': (files['template'], 'template.j2'),
            }

            response = c.post('/params', content_type='multipart/form-data', data=data,
                              headers={'Authorization': 'token'})

            assert response.status_code == 200
            assert response.is_json
            assert 'keys' in response.json
            assert response.json['keys'] == ['key', 'template']

    def test_parameterized(self):
        app = ParamMS()
        with app.test_client() as c:
            response = c.get('/123', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == '123'
            response = c.get('/concat/123/456', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == '123456'
            response = c.get('/value', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == '123'
            response = c.put("/value", json={'value': "456"}, headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == '456'
            response = c.get("/value", headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == '456'
            response = c.get('/param/test1', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'test1default1default2'
            response = c.get('/param/test1?param1=value1', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'test1value1default2'
            response = c.get('/param/test1?param2=value2', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'test1default1value2'
            response = c.get('/param/test1?param1=value1&param2=value2', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'test1value1value2'
            response = c.get('/param/test1?param1=value1&param1=value2', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == "test1['value1', 'value2']default2"

    def test_slug_parameterized(self):
        app = ParamMS()
        with app.test_client() as c:
            response = c.get('/123', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == '123'
            response = c.get('/concat/123/456', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == '123456'

    def test_tuple_returned(self):
        app = TupleReturnedMS()
        with app.test_client() as c:
            headers = {'Accept': 'text/plain', 'Authorization': 'token'}
            response = c.get('/', headers=headers)
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'ok'
            assert response.headers['content-type'] == 'application/json'
            response = c.get('/json', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.json['value'] == 'ok'
            assert response.headers['content-type'] == 'application/json'
            response = c.get('/resp/ok', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == 'ok'
            assert response.headers['content-type'] == 'application/json'
            response = c.get('/tuple/test', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.headers['content-type'] == 'application/json'
            assert response.headers['x-test'] == 'true'
            assert response.get_data(as_text=True) == 'test'

    def test_entry_not_unique(self):
        app = AmbiguousMS()
        with app.test_request_context():
            assert '/{uid}' in app.routes
            assert '/test' in app.routes

    def test_call_not_unique(self):
        app = AmbiguousMS()
        with app.test_client() as c:
            response = c.get('/123', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.get_data(as_text=True) == '123'
            response = c.get('/test', headers={'Authorization': 'token'})
            assert response.status_code == 200
            assert response.json == {'value': "ok"}



import warnings

from coworks.biz.sensors import *  # legacy

warnings.warn(
    "This module is deprecated. Please use `coworks.biz` or `coworks.biz.sensors` instead.",
    DeprecationWarning,
    stacklevel=2,
)



import base64
import io
import re
import typing as t
import urllib
import urllib.parse
from io import BytesIO

from flask import Request as FlaskRequest
from flask import Response as FlaskResponse
from flask import current_app
from flask import json
from requests_toolbelt.multipart.decoder import MultipartDecoder
from werkzeug.datastructures import ETags
from werkzeug.datastructures import FileStorage
from werkzeug.datastructures import Headers
from werkzeug.datastructures import MultiDict
from werkzeug.exceptions import BadRequest
from werkzeug.exceptions import HTTPException
from werkzeug.exceptions import MethodNotAllowed
from werkzeug.exceptions import NotFound
from werkzeug.routing import MapAdapter


class TokenResponse:
    """AWS authorization response."""

    def __init__(self, allow: str | bool, arn: str):
        """Value may be string when allowed only if match workspace label."""
        self.allow = allow
        self.arn = arn

    @property
    def json(self) -> dict:
        return {
            "principalId": "user",
            "policyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Action": "execute-api:Invoke",
                        "Effect": "Allow" if self.allow else "Deny",
                        "Resource": self.arn
                    }
                ]
            }
        }


class CoworksMapAdapter(MapAdapter):

    def __init__(self, environ, url_map, aws_url_map, stage_prefixed):
        server_name = environ["SERVER_NAME"]
        aws_stage = environ["aws_stage"]
        url_scheme = environ["REQUEST_SCHEME"]
        path_info = environ["PATH_INFO"]
        method = environ["REQUEST_METHOD"]
        script_name = f'/{aws_stage}/' if stage_prefixed else ''
        super().__init__(url_map, server_name=server_name, script_name=script_name, subdomain='',
                         url_scheme=url_scheme, path_info=path_info, default_method=method)
        self.aws_url_map = aws_url_map
        self.aws_entry_path = environ["aws_entry_path"]
        self.aws_entry_path_parameters = environ["aws_entry_path_parameters"]

    def match(self, method=None, return_rule=False, **kwargs):
        try:
            if self.aws_entry_path not in self.aws_url_map:
                raise NotFound()

            rules = self.aws_url_map[self.aws_entry_path]
            for rule in rules:
                if self.default_method in rule.methods:
                    return rule if return_rule else rule.endpoint, self.aws_entry_path_parameters
            else:
                raise MethodNotAllowed()
        except HTTPException:
            raise
        except Exception as e:
            current_app.logger.debug(f"Rule match error {e}")
            raise NotFound()


class CoworksResponse(FlaskResponse):
    """Default mimetype is redefined."""
    default_mimetype = "application/json"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CoworksRequest(FlaskRequest):

    def __init__(self, environ, **kwargs):
        self.aws_event: dict | None = environ.get('aws_event')
        self.aws_context = environ.get('aws_context')
        self.aws_query_string: MultiDict | None = environ.get('aws_query_string')
        self.aws_body: str | bytes | None = environ.get('aws_body')
        self._in_lambda_context: bool = self.aws_event is not None
        self.__stream: BytesIO | None = None
        self.__form: MultiDict[str, str] | None = None
        self.__files: MultiDict | FileStorage | None = None
        self.__data = self.__json = None

        super().__init__(environ, **kwargs)

        if self.aws_event and self._in_lambda_context:
            self.headers = Headers(self.aws_event.get('headers'))

    @property
    def in_lambda_context(self):
        """Defined as a property to be read only."""
        return self._in_lambda_context

    @property
    def is_json(self) -> bool:
        """If no content type defined in request, default is application/json`.        """
        return not self.mimetype or super().is_json

    @property
    def is_multipart(self) -> bool:
        """Check if the mimetype indicates form-data.
        """
        mt = self.mimetype
        return (
                mt == "multipart/form-data"
        )

    @property
    def is_form_urlencoded(self) -> bool:
        """Check if the mimetype indicates form-data.
        """
        mt = self.mimetype
        return (
                mt == "application/x-www-form-urlencoded"
        )

    @property
    def args(self):
        if not self.in_lambda_context:
            return super().args
        return self.aws_query_string

    @property
    def stream(self):
        if not self.in_lambda_context:
            return super().stream

        if self.__stream is None:
            self._load_stream_form_files()
        return self.__stream

    @property
    def form(self):
        if not self.in_lambda_context:
            return super().form

        if self.__form is None:
            self._load_stream_form_files()
        return self.__form

    @property
    def files(self):
        if not self.in_lambda_context:
            return super().files

        if self.__files is None:
            self._load_stream_form_files()
        return self.__files

    def get_data(self, **kwargs):
        if not self.in_lambda_context:
            return super().get_data(**kwargs)

        if self.__data is None:
            if kwargs.get('as_text', False):
                self.__data = json.dumps(self.aws_body) if self._body_is_dict else self.aws_body
            elif self.aws_body:
                self.__data = self.aws_body if self._body_is_dict else json.loads(self.aws_body)
        return self.__data

    def get_json(self, **kwargs):
        if not self.in_lambda_context:
            return super().get_json(**kwargs)

        if self.__json is None and self.aws_body:
            self.__json = self.aws_body if self._body_is_dict else json.loads(self.aws_body)
        return self.__json

    @property
    def if_match(self):  # No cache
        return ETags()

    @property
    def if_none_match(self):  # No cache
        return ETags()

    @property
    def if_modified_since(self):  # No cache
        return None

    @property
    def _body_is_dict(self) -> bool:
        return isinstance(self.aws_body, dict)

    def _load_stream_form_files(self) -> None:

        # Stream part
        if self.is_multipart and isinstance(self.aws_body, str):
            self.__stream = io.BytesIO(base64.b64decode(self.aws_body))
        elif self.is_form_urlencoded and isinstance(self.aws_body, str):
            self.__stream = io.BytesIO(self.aws_body.encode('ascii'))
        else:
            raise BadRequest(f'Undefined mime-type for stream body: {self.mimetype}')

        if self.is_multipart:
            self.__files = MultiDict()
            self.__form = MultiDict()
            multipart_data = MultipartDecoder(self.aws_body, self.content_type)
            for part in multipart_data.parts:

                # Files part
                if b'content-disposition' in part.headers:
                    content_disposition = part.headers.get(b'content-disposition').decode("utf-8")
                    filename_regexp = "filename=\"(?P<filename>[^\"]+)\""
                    match = re.search(filename_regexp, content_disposition)
                    filename = match.group('filename') if match else None

                    name_regexp = "name=\"(?P<name>[^\"]+)\""
                    match = re.search(name_regexp, content_disposition)
                    name = match.group('name') if match else None

                    if b'content-type' in part.headers:
                        content_type = part.headers.get(b'content-type').decode("utf-8")
                    else:
                        content_type = None

                    if b'content-length' in part.headers:
                        content_length = part.headers.get(b'content-length').decode("utf-8")
                    else:
                        content_length = None

                    self.__files[filename] = FileStorage(
                        stream=io.BytesIO(part.content),
                        filename=filename,
                        name=name,
                        content_type=content_type,
                        content_length=content_length,
                    )

                # Form part

        elif self.is_form_urlencoded:
            self.__files = MultiDict()
            self.__form = MultiDict(urllib.parse.parse_qs(t.cast(str, self.aws_body)))



import os
from unittest import mock

from flask import json

from coworks import Blueprint
from coworks import TechMicroService
from coworks import entry
from coworks.blueprint.admin_blueprint import Admin


class DocumentedMS(TechMicroService):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.register_blueprint(Admin(), url_prefix="/admin")

    def token_authorizer(self, token):
        return True

    @entry
    def get(self):
        """Root access."""
        return "get"

    @entry
    def post(self):
        """Root access."""
        return "post"

    @entry
    def post_content(self, value, other="none"):
        """Add content."""
        return f"post_content {value}{other}"

    @entry
    def post_contentannotated(self, value: int, other: str = "none"):
        """Add content."""
        return f"post_content {value}{other}"

    @entry(no_auth=True)
    def get_list(self, values: [int]):
        return "ok"


class HiddenBlueprint(Blueprint):

    @entry
    def get(self):
        """Test not in routes."""
        return "ok"


@mock.patch.dict(os.environ, {"TOKEN": "token"})
class TestClass:

    def test_routes(self):
        app = DocumentedMS()
        with app.test_request_context():
            assert '/' in app.routes
            assert '/content/{value}' in app.routes
            assert '/contentannotated/{value}' in app.routes
            assert '/list/{values}' in app.routes

    def test_documentation(self):
        app = DocumentedMS()
        with app.test_client() as c:
            response = c.get('/admin/route?blueprint=test', headers={'Authorization': 'token'})
            assert response.status_code == 404

            response = c.get('/admin/route?blueprint=__all__', headers={'Authorization': 'token'})
            assert response.status_code == 200
            routes = json.loads(response.get_data(as_text=True))
            assert routes["/"]['GET'] == {
                "doc": "Root access.", "signature": "()", "endpoint": "get",
                'binary_headers': None, 'no_auth': False, 'no_cors': True
            }
            assert routes["/"]['POST'] == {
                "doc": "Root access.", "signature": "()", "endpoint": "post",
                'binary_headers': None, 'no_auth': False, 'no_cors': True
            }
            assert routes["/content/<value>"]['POST'] == {
                "doc": "Add content.", "signature": "(value, other=none)", "endpoint": "post_content",
                'binary_headers': None, 'no_auth': False, 'no_cors': True
            }
            assert routes["/contentannotated/<value>"]['POST'] == {
                "doc": "Add content.", "signature": "(value:<class 'int'>, other:<class 'str'>=none)",
                'endpoint': "post_contentannotated",
                'binary_headers': None, 'no_auth': False, 'no_cors': True
            }
            assert routes["/admin/route"]['GET']['signature'] == "(prefix=None, blueprint=None)"
            assert routes["/admin/route"]['GET']['endpoint'] == "admin.get_route"
            assert routes["/list/<values>"]['GET'] == {
                "signature": "(values:[<class 'int'>])", 'endpoint': 'get_list',
                'binary_headers': None, 'no_auth': True, 'no_cors': True
            }

            response = c.get('/admin/route', headers={'Authorization': 'token'})
            assert response.status_code == 200
            routes = json.loads(response.get_data(as_text=True))
            assert "/admin/routes" not in routes

    def test_documentation_with_hidden_blueprints(self):
        app = DocumentedMS()
        app.register_blueprint(HiddenBlueprint(), url_prefix="/hidden", hide_routes=True)
        with app.test_client() as c:
            response = c.get('/admin/route', headers={'Authorization': 'token'})
            assert response.status_code == 200
            routes = json.loads(response.get_data(as_text=True))
            assert '/hidden' not in routes



import os

from coworks import TechMicroService
from coworks import entry
from coworks.blueprint.admin_blueprint import Admin
from coworks.blueprint.profiler_blueprint import Profiler


# from aws_xray_sdk.core import xray_recorder
# from coworks.extension.xray import XRay


class MyMicroService(TechMicroService):
    DOC_MD = """
## Microservice for ...
"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.register_blueprint(Admin(), url_prefix='/admin')
        if os.getenv('CWS_STAGE') == "dev":
            self.register_blueprint(Profiler(self), url_prefix='/profiler')

        @self.before_request
        def before():
            ...

        @self.errorhandler(500)
        def handle(e):
            ...

    def init_cli(self):
        """
        from flask_migrate import Migrate
        Migrate(self, db)
        """

    def token_authorizer(self, token):
        # Simple authorization process.
        # If you want to access AWS event or context for a more complex case, override the function _token_handler.
        return token in os.getenv('USER_KEYS', '').split(',')

    @entry
    def get(self):
        return 'project ready!\n'


app = MyMicroService()

# For this extension you need to install 'aws_xray_sdk'.
# XRay(app, xray_recorder)



import typing as t

from .coworks import Blueprint
from .coworks import TechMicroService
from .coworks import entry
from .globals import request
from .version import __version__

_all__ = (
    TechMicroService, Blueprint, entry,
    request,
    __version__
)


def __getattr__(name: str) -> t.Any:
    if name == "__version__":
        import importlib.metadata
        import warnings

        warnings.warn(
            "The '__version__' attribute is deprecated and will be removed.",
            DeprecationWarning,
            stacklevel=2,
        )
        return importlib.metadata.version("coworks")

    raise AttributeError(name)



import pytest

from coworks import entry
from tests.coworks.blueprint.blueprint import BP
from tests.coworks.ms import SimpleMS


def get_event(token):
    return {
        'type': 'TOKEN',
        'methodArn': 'arn:aws:execute-api:eu-west-1:935392763270:htzd2rneg1/dev/GET/',
        'authorizationToken': token
    }


class AuthorizeAll(SimpleMS):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.register_blueprint(BP(), url_prefix="/blueprint")

    def token_authorizer(self, token):
        return True

    @entry
    def get_product(self, ref):
        return ref


class AuthorizeNothing(AuthorizeAll):

    def token_authorizer(self, token):
        return False


class AuthorizedMS(AuthorizeAll):

    def token_authorizer(self, token):
        return token == 'token'


class AuthorizeExceptionMS(AuthorizeAll):

    def token_authorizer(self, token):
        raise Exception()


class TestClass:

    def test_authorize_all(self, empty_aws_context):
        app = AuthorizeAll()
        with app.app_context() as c:
            response = app(get_event('/token'), empty_aws_context)
            assert response['principalId'] == 'user'
            assert response['policyDocument']['Statement'][0]['Effect'] == 'Allow'

    def test_authorize_nothing(self, empty_aws_context):
        app = AuthorizeNothing()
        with app.app_context() as c:
            response = app(get_event('token'), empty_aws_context)
            assert response['principalId'] == 'user'
            assert response['policyDocument']['Statement'][0]['Effect'] == 'Deny'

    def test_authorized(self, empty_aws_context):
        app = AuthorizedMS()
        with app.app_context() as c:
            response = app(get_event('wrong'), empty_aws_context)
            assert response['principalId'] == 'user'
            assert response['policyDocument']['Statement'][0]['Effect'] == 'Deny'
            response = app(get_event('token'), empty_aws_context)
            assert response['principalId'] == 'user'
            assert response['policyDocument']['Statement'][0]['Effect'] == 'Allow'

    def test_authorize_exception(self, empty_aws_context):
        app = AuthorizeExceptionMS()
        with app.app_context() as c:
            response = app(get_event('token'), empty_aws_context)
            assert response['principalId'] == 'user'
            assert response['policyDocument']['Statement'][0]['Effect'] == 'Deny'



import os
from unittest import mock

import pytest
from click import UsageError

from coworks.cws.client import client


class TestClass:
    @mock.patch.dict(os.environ, {"FLASK_APP": "command:app"})
    def test_no_project_file_no_module(self, capsys):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['--project-dir', '.', 'test'], obj={})
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0

    @mock.patch.dict(os.environ, {"FLASK_APP": "command:app"})
    def test_wrong_project_dir(self, example_dir):
        with pytest.raises(UsageError) as pytest_wrapped_e:
            client.main(['--project-dir', 'doesntexist', 'test'], 'cws', standalone_mode=False)
        msg = "Project dir /home/gdo/workspace/coworks/tests/cws/src/doesntexist not found."
        assert pytest_wrapped_e.value.args[0] == msg
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            client(prog_name='cws', args=['-p', 'doesntexist', 'test'])

    def test_wrong_project_config_version(self, example_dir):
        with pytest.raises(RuntimeError) as pytest_wrapped_e:
            client.main(['--project-dir', '.', '--config-file-suffix', '.wrong.yml', 'cmd'], 'cws',
                        standalone_mode=False)
        assert pytest_wrapped_e.type == RuntimeError
        assert pytest_wrapped_e.value.args[0] == "Wrong project file version (should be 3).\n"



import io
from unittest.mock import MagicMock

from coworks import TechMicroService
from coworks import entry
from coworks import request


class TechMS(TechMicroService):
    def __init__(self, **kwargs):
        super().__init__('test', **kwargs)

    def token_authorizer(self, token):
        return True


class S3MockTechMS(TechMicroService):
    def __init__(self):
        super().__init__('test')
        session = MagicMock()
        session.client = MagicMock()
        s3_object = {'Body': io.BytesIO(b'test'), 'ContentType': 'text/plain'}
        session.client.get_object = MagicMock(return_value=s3_object)
        self.aws_s3_run_session = session


class SimpleMS(TechMicroService):

    @entry
    def get(self):
        """Root access."""
        return "get"

    def get1(self):
        """Not recognized."""
        return "get1"

    @entry
    def get_content(self):
        return "get content"

    @entry
    def get__content(self, value):
        return f"get content with {value}"

    @entry
    def get_content__(self, value, other):
        return f"get content with {value} and {other}"

    @entry
    def post_content(self, other="none"):
        return f"post content without value but {other}"

    @entry
    def post_content_(self, value, other="none"):
        return f"post content with {value} and {other}"

    # **param
    @entry
    def get_kwparam1(self, value=0):
        return f"get **param with only {value}"

    @entry
    def get_kwparam2(self, value=0, **kwargs: dict):
        return f"get **param with {value} and {list(kwargs.keys())}"

    @entry
    def put_kwparam2(self, value=0, **kwargs):
        return f"get **param with {value} and {list(kwargs.keys())}"

    @entry
    def post_kwparam2(self, value, **kwargs):
        return f"get **param with {value} and {list(kwargs.keys())}"

    # composed path
    @entry
    def get_extended_content(self):
        return "hello world"

    @entry
    def get_dir_content(self):
        return {'msg': "hello world"}


class GlobalMS(TechMicroService):

    def token_authorizer(self, token):
        return True

    @entry
    def get_event_method(self):
        return request.aws_event['httpMethod']



import base64
import datetime
import os
import re
import secrets
import string
import typing as t

import boto3
import requests
from Crypto.Cipher import AES
from flask import json
from flask import current_app, request
from werkzeug.exceptions import BadRequest
from werkzeug.exceptions import Forbidden
from werkzeug.exceptions import MethodNotAllowed
from werkzeug.exceptions import NotFound

from coworks import TechMicroService
from coworks import entry
from coworks.utils import is_json


class DirectoryMicroService(TechMicroService):
    DOC_MD = """
## Directory service

Microservice to get information on a deployed CoWorks TechMicroServices.

You can also call a microservice by its name or get a code for a temporary authorized call.

See 'samples/directory' to get how to create and deploy it.
"""

    def __init__(self, name="directory", session=None, access_key_var="AWS_USER_ACCESS_KEY_ID",
                 secret_access_key_var="AWS_USER_SECRET_ACCESS_KEY", region_name_var="AWS_REGION_NAME", **kwargs):
        super().__init__(name=name, **kwargs)

        if session is None:
            access_key = os.getenv(access_key_var)
            secret_key = os.getenv(secret_access_key_var)
            region_name = os.getenv(region_name_var)
            session = boto3.Session(access_key, secret_key, region_name=region_name)
        self.api_client = session.client('apigateway')
        self.lambda_client = session.client('lambda')

    @entry
    def post(self, pattern=None, position=None):
        """Get the list of microservices matching a pattern.

        :param pattern: AWS name pattern (ex. ".*directory.*").
        :param position: position for pagination.
        """
        try:
            match_re = re.compile(pattern) if pattern else None
        except re.error as e:
            raise BadRequest(f"Wrong regular expression: {e.msg}")

        def found(a):
            return match_re.search(a['name']) if match_re else True

        boto3_args = {'limit': 100}
        if position:
            boto3_args.update({'position': position})
        try:
            objects = self.api_client.get_rest_apis(**boto3_args)
        except (Exception,) as e:
            raise BadRequest(e.msg)
        api = [*filter(found, objects['items'])]
        position = self._get_position(objects)
        while position and len(api) < 500:
            objects = self.api_client.get_rest_apis(position=position, limit=100)
            api = [*api, *filter(found, objects['items'])]
            position = self._get_position(objects)

        return {
            'items': {a['name']: a['id'] for a in api},
            'position': position if position else ''
        }

    @entry
    def get_aws(self, name):
        """Get AWS information on a microservice.

        :param name: AWS microservice's name.
        """

        def found(apis):
            for a in apis:
                if a['name'] == name:
                    return a

        objects = self.api_client.get_rest_apis(limit=100)
        api = found(objects['items'])
        if api:
            return api
        position = objects['position']
        while position:
            objects = self.api_client.get_rest_apis(position=position, limit=100)
            api = found(objects['items'])
            if api:
                return api
            position = objects['position'] if 'position' in objects else None

        raise NotFound()

    @entry
    def get_stages(self, name):
        """Get microservice's stages for the microservice.

        :param name: AWS microservice's name.
        """
        api = self.get_aws(name)
        api_id = api['id']
        stages = self.api_client.get_stages(restApiId=api_id)
        return {s['stageName']: {**s, 'api_id': api_id} for s in stages['item']}

    @entry
    def get_url(self, name, stage=None, token_var_name='TOKEN'):
        """Get microservice url for a microservice.
        If stage is not defined the latest production version is choosen or 'dev' if no production version.

        :param name: AWS microservice's name.
        :param stage: AWS microservice's stage.
        :param token_var_name: lambda environment variable's name for the token.
        """
        api_id = ""
        stages = self.get_stages(name)
        if stage:
            if stage not in stages:
                raise NotFound()
        else:
            stage = sorted(stages)[-1]
        api_id = stages[stage]['api_id']

        auths = self.api_client.get_authorizers(restApiId=api_id)
        lambda_uri = auths['items'][0]['authorizerUri']
        uris = lambda_uri.split(':')
        lambda_name = f"{uris[-1].split('$')[0]}{stage}"
        lambda_fun = self.lambda_client.get_function(FunctionName=lambda_name)

        url = f"https://{api_id}.execute-api.eu-west-1.amazonaws.com/{stage}"
        token = lambda_fun['Configuration']['Environment']['Variables'].get(token_var_name)
        return {'url': url, 'token': token}

    @entry
    def get_doc(self, name, stage=None):
        """Get microservice documentation.
        If stage is not defined the latest production version is choosen or 'dev' if no production version.

        :param name: AWS microservice's name.
        :param stage: AWS microservice's stage.
        """
        info = self.get_url(name, stage=stage)
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'text/html; charset utf-8',
        }
        url = f"{info['url']}/admin"
        res = requests.get(url, headers=headers)
        return res.text, res.status_code, dict(res.headers)

    @entry
    def get_name(self, api_id):
        """Get microservice name from api id.

        :param api_id: AWS API ID.
        """
        return self.api_client.get_rest_api(restApiId=api_id)

    @entry
    def post_call(self, name, stage=None, path='/admin', method='get', token_var_name='TOKEN', data=None):
        """Call a microservice from its name.

        Accept and Content-Type may be defined in header for the microservice call.

        :param name: microservice's name.
        :param stage: stage version.
        If stage is not defined the latest production version is choosen or 'dev' if no production version.
        :param path: entry path (default '/admin').
        :param method: method called (default 'get').
        :param token_var_name: token variable name (default 'TOKEN').
        :param data: query parameters or json body (default {}).
        """
        data = data or {}
        info = self.get_url(name, stage=stage, token_var_name=token_var_name)

        headers = {
            'Accept': request.headers.get('Accept', 'application/json'),
            'Authorization': info['token'],
            'Content-Type': request.headers.get('Content-Type', 'application/json'),
        }

        url = f"{info['url']}{path}"

        current_app.logger.info(f"{method.upper()} on {url}, with {data}")
        if method.upper() == 'GET':
            res = requests.get(url, params=data, headers=headers)
        elif method.upper() == 'POST':
            res = requests.post(url, json=data, headers=headers)
        else:
            raise MethodNotAllowed()

        if not res.ok:
            return res.text, res.status_code

        accept = headers.get('Accept', 'application/json')
        return res.json() if is_json(accept) else res.text, res.status_code, dict(res.headers)

    @entry
    def post_code(self, name, duration=60, **kwargs):
        """Returns a temporary code to call a microservice.

        :param name: microservice's name.
        :param duration: token duration in minutes.
        :param kwargs: call parameters.
        """
        kwargs['name'] = name

        # Adds expiration time
        expiration_time = datetime.datetime.now() + datetime.timedelta(minutes=duration)
        kwargs['expiration_time'] = expiration_time.strftime('%Y-%m-%dT%H:%M:%S%z')

        # Encodes invocation data
        iv = ''.join((secrets.choice(string.ascii_letters) for i in range(AES.block_size)))
        crypto_obj = AES.new(self._pad(os.getenv("TOKEN")).encode('utf-8'), AES.MODE_CBC, iv.encode('utf-8'))
        cipher = crypto_obj.encrypt(self._pad(json.dumps(kwargs), ' ').encode('utf-8'))
        return iv + base64.b64encode(cipher).decode()

    @entry
    def post_decode(self, code=None):
        """Call the microservice defined by the code with GET method."""
        return self._decode(code=code)

    @entry(no_auth=True)
    def post_code_call(self, code=None, data=None):
        """Call the microservice defined by the code with POST method."""
        return self._call(code=code, data=data)

    def _decode(self, code=None):
        data = {}

        # Decodes invocation data
        iv = code[:AES.block_size]
        code = base64.b64decode(code[AES.block_size:])
        crypto_obj = AES.new(self._pad(os.getenv("TOKEN")).encode('utf-8'), AES.MODE_CBC, iv.encode('utf-8'))
        cipher = crypto_obj.decrypt(code).decode('ascii')
        call_data = json.loads(cipher)
        current_app.logger.debug(call_data)

        # Checks expiration time
        expiration_time = datetime.datetime.strptime(call_data.pop('expiration_time'), '%Y-%m-%dT%H:%M:%S')
        now = datetime.datetime.now()
        if now > expiration_time:
            current_app.logger.debug(f"Expired call : {now} > {expiration_time}")
            raise Forbidden("Your link is no more available.")

        # Check path prefix
        if 'path' in call_data and 'path' in data:
            if call_data['path'] not in data['path']:
                current_app.logger.debug(f"{call_data['path']} not in {data.get('path')}")
                raise Forbidden("You cannot access this entry.")

        # Checks mandatory parameters
        return call_data

    def _call(self, code=None, data=None):
        data = data or {}
        if 'name' in data:
            raise BadRequest("Cannot overload name parameters")

        final_data = {**data, **self._decode(code)}
        if 'path' not in final_data:
            raise BadRequest("Missing path parameters")

        # Invokes microservice entry
        return self.post_call(**final_data)

    def _get_position(self, objects):
        return objects['position'] if 'position' in objects else None

    def _pad(self, str: str, padding_char: t.Optional[str] = None) -> str:
        """Pads the string to AES.block_size (pad with padding_char)"""
        padding = (AES.block_size - (len(str) % AES.block_size))
        return str + padding * (padding_char if padding_char is not None else chr(padding))



import typing as t

from pydantic import BaseModel
from pydantic import ConfigDict

from .data import CursorPagination
from .data import JsonApiDataMixin


class Pagination(t.Protocol, t.Iterable):
    total: int
    page: int
    pages: int
    per_page: int
    has_prev: bool
    prev_num: int | None
    has_next: bool
    next_num: int | None


@t.runtime_checkable
class Query(t.Protocol):
    def paginate(self, *, page, per_page, max_per_page) -> Pagination:
        ...

    def all(self) -> list[JsonApiDataMixin]:
        ...


class ListPagination(CursorPagination):
    total: int = 0
    values: list[t.Any]

    def __init__(self, **data):
        super().__init__(**data)
        self.total = len(self.values)

    def __iter__(self):
        assert self.page is not None  # by the validator
        assert self.per_page is not None  # by the validator
        return self.values[(self.page - 1) * self.per_page: self.page * self.per_page].__iter__()


class ListQuery(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    values: list[JsonApiDataMixin]

    def paginate(self, *, page, per_page, max_per_page) -> Pagination:
        return ListPagination(values=self.values, page=page, per_page=per_page)  # type: ignore[return-value]

    def all(self) -> list[JsonApiDataMixin]:
        return self.values



import os
import smtplib
import typing as t
from email import message
from email.utils import formataddr
from email.utils import formatdate
from email.utils import make_msgid

import requests
from flask import current_app
from flask import render_template_string
from werkzeug.datastructures import FileStorage
from werkzeug.exceptions import BadRequest
from werkzeug.exceptions import InternalServerError

from coworks import Blueprint
from coworks import entry


#
# BLUEPRINT PART
#

class Mail(Blueprint):
    """Mail blueprint.
    Initialization parameters must be in environment (Twelve-Factor App).
    Environment variables needed:
    - env_server_var_name: Variable name for the SMTP server.
    - env_port_var_name: Variable name for the SMPT port.
    - env_login_var_name: Variable name for the login.
    - env_passwd_var_name: Variable name for the password.
    """

    def __init__(self, name: str = "mail",
                 env_server_var_name: str = '', env_port_var_name: str = '',
                 env_login_var_name: str = '', env_passwd_var_name: str = '',
                 env_var_prefix: str = '', **kwargs: dict):
        super().__init__(name=name, **kwargs)

        # global prefixed veriables
        if env_var_prefix:
            env_server_var_name = f"{env_var_prefix}_SERVER"
            env_port_var_name = f"{env_var_prefix}_PORT"
            env_login_var_name = f"{env_var_prefix}_LOGIN"
            env_passwd_var_name = f"{env_var_prefix}_PASSWD"

        self.smtp_server = os.getenv(env_server_var_name)
        if not self.smtp_server:
            raise RuntimeError(f'{env_server_var_name} not defined in environment.')
        self.smtp_port = int(os.getenv(env_port_var_name, 587))
        self.smtp_login = os.getenv(env_login_var_name)
        if not self.smtp_login:
            raise RuntimeError(f'{env_login_var_name} not defined in environment.')
        self.smtp_passwd = os.getenv(env_passwd_var_name)
        if not self.smtp_passwd:
            raise RuntimeError(f'{env_passwd_var_name} not defined in environment.')

    @entry
    def post_send(self, subject: str = "", from_addr: str | None = None, from_name: str = '',
                  reply_to: str | None = None, body: str = "", body_template: str | None = None, body_type="plain",
                  to_addrs: list[str] | None = None, cc_addrs: list[str] | None = None,
                  bcc_addrs: list[str] | None = None,
                  attachments: FileStorage | t.Iterator[FileStorage] | None = None, attachment_urls: dict | None = None,
                  starttls: bool = True, data: dict | None = None):
        """ Send mail.
        To send attachments, add files in the body of the request as multipart/form-data.

        :param subject: Email's subject (required).
        :param from_addr: From recipient.
        :param from_name: Name besides from_address.
        :param reply_to: Reply recipient.
        :param to_addrs: Email's recipients. Accept one or several email addresses separated by commas.
        :param cc_addrs: Email's cc recipients. Accept one or several email addresses separated by commas.
        :param bcc_addrs: Email's bcc recipients. Accept one or several email addresses separated by commas.
        :param body: Email's body (required if body_template not defined).
        :param body_template: Email's body template (required if body not defined).
        :param body_type: Email's body type.
        :param attachments: File storage.
        :param attachment_urls: File url.
        :param starttls: Puts the connection to the SMTP server into TLS mode.
        :param data: Jinnja context for the templating.
        """

        from_addr = from_addr or os.getenv('from_addr')
        if not from_addr:
            raise BadRequest("From address not defined (from_addr:str)")
        to_addrs = to_addrs or os.getenv('to_addrs')
        if not to_addrs:
            raise BadRequest("To addresses not defined (to_addrs:[str])")

        if body_template is not None:
            if body is not None:
                raise BadRequest("Body and body_template parameters both defined.")
            body = render_template_string(body_template, **data)

        # Creates email
        try:
            from_ = formataddr((from_name if from_name else False, from_addr))
            msg = message.EmailMessage()
            msg["Date"] = formatdate(localtime=True)
            msg["Message-ID"] = make_msgid()
            msg['Subject'] = subject
            msg['From'] = from_
            if reply_to is not None:
                msg['Reply-To'] = reply_to
            msg['To'] = to_addrs if isinstance(to_addrs, str) else ', '.join(to_addrs)
            if cc_addrs:
                msg['Cc'] = cc_addrs if isinstance(cc_addrs, str) else ', '.join(cc_addrs)
            if bcc_addrs:
                msg['Bcc'] = bcc_addrs if isinstance(bcc_addrs, str) else ', '.join(bcc_addrs)
            msg.set_content(body, subtype=body_type)

            if attachments:
                current_app.logger.info("Attachments defined")
                if not isinstance(attachments, list):
                    attachments = [attachments]
                for attachment in attachments:
                    msg.add_attachment(attachment.stream.read(), maintype='multipart', subtype=attachment.content_type)
                    current_app.logger.info(f"Add attachment {attachment}")

            if attachment_urls:
                current_app.logger.info("Attachments urls defined")
                for attachment_name, attachment_url in attachment_urls.items():
                    response = requests.get(attachment_url)
                    if response.status_code == 200:
                        attachment = response.content
                        maintype, subtype = response.headers['Content-Type'].split('/')
                        msg.add_attachment(attachment, maintype=maintype, subtype=subtype,
                                           filename=attachment_name)
                        current_app.logger.info(f"Add attachment {attachment_name} - size {len(attachment)}")
                    else:
                        return f"Failed to download attachment, error {response.status_code}.", 400

        except Exception as e:
            raise BadRequest(f"Cannot create email message (Error: {str(e)}).")

        # Send email
        try:
            with smtplib.SMTP(self.smtp_server, port=self.smtp_port) as server:
                if starttls:
                    server.starttls()
                server.login(self.smtp_login, self.smtp_passwd)
                server.send_message(msg)

            resp = f"Mail sent to {msg['To']}"
            current_app.logger.info(resp)
            return resp
        except smtplib.SMTPAuthenticationError:
            raise ConnectionError("Wrong username/password : cannot connect.")
        except Exception as e:
            raise InternalServerError(f"Cannot send email message (Error: {str(e)}).")



import importlib
import os
import sys
import types
import typing as t
from contextlib import contextmanager
from logging import WARNING
from logging import getLogger
from pathlib import Path

import click
import flask
import yaml
from click import UsageError
from flask.cli import ScriptInfo

from coworks import __version__
from coworks.utils import DEFAULT_PROJECT_DIR
from coworks.utils import PROJECT_CONFIG_VERSION
from .deploy import deploy_command
from .deploy import deployed_command
from .deploy import destroy_command
from .new import new_command
from .utils import get_system_info
from .utils import show_stage_banner


class CwsScriptInfo(ScriptInfo):

    def __init__(self, project_dir=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_dir = project_dir

    @property
    def project_dir(self):
        return self.__project_dir

    @project_dir.setter
    def project_dir(self, project_dir):
        self.__project_dir = Path(project_dir).absolute() if project_dir else None

    @contextmanager
    def project_context(self, ctx=None):
        old_dir = os.getcwd()
        try:
            os.chdir(self.project_dir)
        except OSError:
            if not ctx:
                raise UsageError(f"Project dir {self.project_dir} not found.")
        finally:
            yield
            os.chdir(old_dir)

    def load_app(self):
        with self.project_context():
            return super().load_app()


def _set_stage(ctx, param, value):
    if value is not None:
        os.environ["CWS_STAGE"] = value
        return value


class CwsGroup(flask.cli.FlaskGroup):

    def __init__(self, add_default_commands=True, **extra):
        params = list(extra.pop("params", None) or ())
        extra["add_version_option"] = False
        extra["load_dotenv"] = True
        super().__init__(params=params, **extra)

        if add_default_commands:
            self.add_command(t.cast(click.Command, new_command))
            self.add_command(t.cast(click.Command, deploy_command))
            self.add_command(t.cast(click.Command, destroy_command))
            self.add_command(t.cast(click.Command, deployed_command))

    def make_context(self, info_name, args, parent=None, **kwargs):
        # Get project infos
        script_info = CwsScriptInfo(create_app=self.create_app, set_debug_flag=self.set_debug_flag)
        if 'obj' not in kwargs:
            ctx = super().make_context(info_name, args, obj=script_info, **kwargs)
        else:
            ctx = super().make_context(info_name, args, **kwargs)
        config_file = ctx.params.get('config_file')
        config_file_suffix = ctx.params.get('config_file_suffix')
        project_dir = ctx.params.get('project_dir')
        if project_dir:
            script_info.project_dir = project_dir
            sys.path.insert(0, project_dir)

        # Adds environment variables and defined commands from project file
        project_config = ProjectConfig(project_dir, config_file, config_file_suffix)
        with script_info.project_context(ctx):
            commands = project_config.get_commands(ctx.params.get('stage'))
            if commands:
                for name, options in commands.items():
                    cmd_class_name = options.pop('class', None)
                    if cmd_class_name:
                        cmd_module, cmd_class = cmd_class_name.rsplit('.', 1)
                        try:
                            cmd = import_attr(cmd_module, cmd_class)
                        except ModuleNotFoundError:
                            raise click.UsageError(f"Cannot load command {cmd_class!r} in module {cmd_module!r}.")
                    elif name in self.commands:
                        cmd = self.commands[name]
                    else:
                        raise click.UsageError(f"The command {name} is undefined or the class option is missing.")

                    # Sets option's value as default command param
                    # (may then be forced in command line or defined by default)
                    for param in cmd.params:
                        param_name = param.name.replace('-', '_')
                        if param_name in options:
                            param.default = options.get(param_name)

                    self.add_command(cmd, name)

        return ctx


@click.group(cls=CwsGroup)
@click.version_option(version=__version__, message=f'%(prog)s %(version)s, {get_system_info()}')
@click.option('-p', '--project-dir', default=DEFAULT_PROJECT_DIR,
              help=f"The project directory path (absolute or relative) [default to '{DEFAULT_PROJECT_DIR}'].")
@click.option('-c', '--config-file', default='project', help="Configuration file path [relative from project dir].")
@click.option('--config-file-suffix', default='.cws.yml', help="Configuration file suffix.")
@click.option('-s', '--stage', default='dev', help="Stage environment.")
def client(*args, **kwargs):
    os.environ['CWS_STAGE'] = kwargs['stage']


class ProjectConfig:
    """Class for the project configuration file."""

    def __init__(self, project_dir, file_name, file_suffix):
        getLogger('anyconfig').setLevel(WARNING)
        self.project_dir = project_dir
        try:
            self.params = self._load_config(project_dir, file_name, file_suffix)
        except TypeError:
            raise RuntimeError(f"Cannot find project coniguration file in {project_dir}")
        if self.params and self.params.get('version', PROJECT_CONFIG_VERSION) != PROJECT_CONFIG_VERSION:
            raise RuntimeError(f"Wrong project file version (should be {PROJECT_CONFIG_VERSION}).\n")

    def get_commands(self, workspace):
        """ Returns the list of commands defined for this microservice."""
        commands = self.params.get('commands', {})

        # Commands may be redefined in the specific workspace
        workspaces = self.params.get('workspaces', {})
        if workspace in workspaces:
            specific_workspace_commands = workspaces[workspace].get('commands', {})
            for cmd, options in specific_workspace_commands.items():
                commands[cmd].update(options)

        return commands

    @staticmethod
    def _load_config(project_dir, file_name, file_suffix):
        """Loads the project configuration file."""

        def load(dir):
            project_dir_path = Path(dir)
            project_file = project_dir_path / (file_name + file_suffix)
            try:
                project_file_content = Path(project_file).read_text()
                return yaml.safe_load(project_file_content)
            except FileNotFoundError:
                return {}

        params = load(project_dir)
        if not params:
            params = load('.')

        if not params:
            click.secho("Warning: no project configuration file found", fg="yellow")

        return params


def overriden_run_banner():
    """Copy the original function and add stage banner."""
    show_server_banner_copy = types.FunctionType(flask.cli.show_server_banner.__code__,
                                                 flask.cli.show_server_banner.__globals__,
                                                 name=flask.cli.show_server_banner.__name__,
                                                 argdefs=flask.cli.show_server_banner.__defaults__,
                                                 closure=flask.cli.show_server_banner.__closure__)

    def show_banner_with_stage(*args):
        show_stage_banner()
        show_server_banner_copy(*args)

    return show_banner_with_stage


# Overrides run banner function to add stage value
flask.cli.show_server_banner = overriden_run_banner()


def import_attr(module, attr: str):
    if not isinstance(attr, str):
        raise AttributeError(f"{attr} is not a string.")
    app_module = importlib.import_module(module)
    if "PYTEST_CURRENT_TEST" in os.environ:
        app_module = importlib.reload(app_module)
    return getattr(app_module, attr)



import typing as t
from math import ceil

from pydantic import BaseModel
from pydantic import field_validator

if t.TYPE_CHECKING:
    from fetching import FetchingContext


class CursorPagination(BaseModel):
    """Pagination based on a cursor model (total and per_page must be defined)"""
    total: int
    page: int | None
    per_page: int | None

    @field_validator("page")
    def set_page(cls, page):
        return page or 1

    @field_validator("per_page")
    def set_per_page(cls, per_page):
        return per_page or 20

    @property
    def pages(self) -> int:
        if not self.total:
            return 1
        assert self.per_page is not None  # by the validator
        return ceil(self.total / self.per_page)

    @property
    def has_prev(self) -> bool:
        assert self.page is not None  # by the validator
        return self.page > 1

    @property
    def prev_num(self) -> int | None:
        if not self.has_prev:
            return None
        assert self.page is not None  # by the validator
        return self.page - 1

    @property
    def has_next(self) -> bool:
        assert self.page is not None  # by the validator
        return self.page < self.pages

    @property
    def next_num(self) -> int | None:
        if not self.has_next:
            return None
        assert self.page is not None  # by the validator
        return self.page + 1


class JsonApiRelationship:
    """Relationship information for jsonapi.
    The id may be given independently of the value.
    """

    def __init__(self, *, type_, id_, value=None):
        self.jsonapi_type = type_
        self.jsonapi_id = id_
        self.value = value

    @property
    def resource_value(self) -> "JsonApiDataMixin":
        return self.value


class JsonApiDataMixin:
    """Any data structure which may be transformed to JSON:API resource.
    """

    @property
    def jsonapi_type(self) -> str:
        return 'unknown'

    @property
    def jsonapi_id(self) -> str:
        return 'unknown'

    @property
    def jsonapi_self_link(self):
        return "https://monsite.com/missing_entry"

    def jsonapi_attributes(self, context: 'FetchingContext', with_relationships: list[str] | None = None) \
            -> tuple[dict[str, t.Any], dict[str, 'JsonApiRelationship']]:
        """Splits the structure in attributes versus relationships."""
        return {}, {}


class JsonApiBaseModel(BaseModel, JsonApiDataMixin):
    """BaseModel data for JSON:API resource"""

    def jsonapi_attributes(self, context: "FetchingContext", with_relationships: list[str] | None = None) \
            -> tuple[dict[str, t.Any], dict[str, 'JsonApiRelationship']]:
        fields = context.field_names(self.jsonapi_type)
        attrs = {k: v for k, v in self.model_dump().items() if (not fields or k in fields)}  # type:ignore
        return attrs, {}


class JsonApiDict(dict, JsonApiDataMixin):
    """Dict data for JSON:API resource"""

    @property
    def jsonapi_type(self) -> str:
        return self['type']

    @property
    def jsonapi_id(self) -> str:
        return str(self['id'])

    def jsonapi_attributes(self, context: "FetchingContext", with_relationships: list[str] | None = None) \
            -> tuple[dict[str, t.Any], dict[str, 'JsonApiRelationship']]:
        fields = context.field_names(self.jsonapi_type)
        attrs = {k: v for k, v in self.items() if (not fields or k in fields)}  # type:ignore
        return attrs, {}



import platform
import sys
import typing as t
from contextlib import contextmanager
from importlib.metadata import version
from threading import Thread
from time import sleep

import click


def get_system_info():
    flask_version = version("flask")

    flask_info = f"flask {flask_version}"
    python_info = f"python {sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}"
    platform_system = platform.system().lower()
    platform_release = platform.release()
    platform_info = f"{platform_system} {platform_release}"
    return f"{flask_info}, {python_info}, {platform_info}"


def show_stage_banner(stage: str = 'dev'):
    click.secho(f" * Stage: {stage}", fg="green")


def show_terraform_banner(cloud: bool, refresh: bool):
    click.secho(f" * Using terraform backend {'cloud' if cloud else 's3'} (refresh={refresh})", fg="green")


class ProgressBar:

    def __init__(self, bar):
        self.bar = bar
        self.stop = False
        self.spin_thread = None

    def echo(self, msg):
        swap = self.bar.format_progress_line
        self.bar.format_progress_line = lambda: msg
        self.bar.render_progress()
        click.echo()
        self.bar.format_progress_line = swap

    def update(self, msg: str | None = None):
        if msg:
            self.echo(msg)
        self.bar.update(1)

    def terminate(self, msg: str | None = None):
        self.stop = True
        if self.spin_thread:
            self.spin_thread.join()
        self.bar.finish()
        self.bar.render_progress()
        if msg:
            self.echo(msg)


class DebugProgressBar:
    def update(self, msg: str):
        if msg:
            click.echo("==> " + msg)

    def echo(self, msg: str):
        if msg:
            click.echo("==> " + msg)


@contextmanager  # type: ignore[arg-type]
def progressbar(length=200, *, label: str, threaded: bool = False) -> t.ContextManager[ProgressBar]:  # type: ignore
    """Spinner progress bar.
    Creates it with a task label and updates it with progress messages using the 'update' function.
    """
    if threaded:
        try:
            with click.progressbar(range(length - 1), label=label.ljust(40), show_eta=False) as bar:
                pb = ProgressBar(bar)

                def display_spinning_cursor():
                    while not pb.stop:
                        sleep(1)
                        if not pb.stop:
                            pb.update()

                spin_thread = Thread(target=display_spinning_cursor)
                spin_thread.start()
                pb.spin_thread = spin_thread
                yield pb
                if not pb.stop:
                    pb.terminate()
        except (Exception,) as e:
            click.echo(f"{type(e).__name__}: {str(e)}")
        finally:
            pb.stop = True
    else:
        yield DebugProgressBar()



import os
import tempfile
from pathlib import PosixPath
from unittest import mock

import boto3
from flask.cli import ScriptInfo

from coworks import Blueprint
from coworks import TechMicroService
from coworks import entry
from coworks.cws.client import CwsScriptInfo
from coworks.cws.deploy import Terraform
from coworks.cws.deploy import TerraformBackend
from coworks.cws.deploy import TerraformContext


class CliCtxMokup:

    def __init__(self, stage='dev'):
        super().__init__()
        self.params = {'project_dir': ".", 'stage': stage}

    def find_root(self):
        return self


class BP(Blueprint):

    @entry
    def get_test(self, index):
        return f"blueprint {self} {index}"

    @entry
    def get_extended_test(self, index):
        return f"blueprint extended test {index}"


class TechMS(TechMicroService):
    def __init__(self, **kwargs):
        super().__init__('test', **kwargs)
        self.register_blueprint(BP())

    def token_authorizer(self, token):
        return True

    @entry
    def get(self):
        return "simple get"

    @entry(binary_headers={'content-type': 'img/webp'}, no_auth=True)
    def get_img(self):
        return b"image content"


class TestClass:

    def test_api_resources(self, example_dir, progressbar):
        app = TechMS()
        with app.test_request_context() as app_ctx:
            info = ScriptInfo(create_app=lambda: app)
            terraform_context = TerraformContext(info, CliCtxMokup(stage='dev'))
            backend = TerraformBackend(terraform_context, None, terraform_dir=".", terraform_refresh=False)
            terraform = Terraform(backend, terraform_dir="terraform", workspace="common")
            api_ressources = terraform.api_resources
        assert len(api_ressources) == 7
        assert api_ressources[''].rules is not None
        assert len(api_ressources[''].rules) == 1
        assert not api_ressources[''].rules[0].cws_binary_headers
        assert not api_ressources[''].rules[0].cws_no_auth
        assert len(api_ressources['img'].rules) == 1
        assert api_ressources['img'].rules[0].cws_binary_headers
        assert api_ressources['img'].rules[0].cws_no_auth
        assert api_ressources['test'].rules is None
        assert api_ressources['test_index'].rules is not None
        assert len(api_ressources['test_index'].rules) == 1
        assert api_ressources['extended'].rules is None

    @mock.patch.dict(os.environ, {"test": "local", "FLASK_RUN_FROM_CLI": "true"})
    def test_deploy_ressources(self, example_dir, progressbar, capsys):
        info = CwsScriptInfo(project_dir='.')
        info.app_import_path = "command:app"
        app = info.load_app()
        with app.test_request_context() as app_ctx:
            info = ScriptInfo(create_app=lambda: app)
            terraform_context = TerraformContext(info, CliCtxMokup(stage='dev'))
            backend = TerraformBackend(terraform_context, None, terraform_dir=".", terraform_refresh=False)
            terraform = Terraform(backend, terraform_dir="terraform", workspace="common")
            api_ressources = terraform.api_resources
        assert len(api_ressources) == 5
        assert '' in api_ressources
        assert 'init' in api_ressources
        assert 'env' in api_ressources
        assert 'value' in api_ressources
        assert 'value_index' in api_ressources
        ter_resource = api_ressources['']
        assert len(ter_resource.rules) == 1
        ter_resource = api_ressources['value_index']
        assert len(ter_resource.rules) == 2

    @mock.patch.dict(os.environ, {"test": "local", "FLASK_RUN_FROM_CLI": "true"})
    def test_deploy_local_cmd(self, monkeypatch, example_dir, progressbar, capsys):
        info = CwsScriptInfo(project_dir='.')
        info.app_import_path = "command:app"
        app = info.load_app()
        with app.test_request_context() as app_ctx:
            options = {
                'project_dir': example_dir,
                'stage': 'dev',
                'timeout': 30,
                'memory_size': 100,
                'terraform_dir': ".",
                'terraform_refresh': False,
            }
            info = ScriptInfo(create_app=lambda: app)
            terraform_context = TerraformContext(info, CliCtxMokup())
            backend = TerraformBackend(terraform_context, None, **options)
            terraform = Terraform(backend, terraform_dir=PosixPath("terraform"), workspace="common")
            with tempfile.NamedTemporaryFile() as fp:
                terraform.generate_file("deploy.j2", fp.name, **options)
                fp.seek(0)
                lines = fp.readlines()
                assert len(lines) == 2049
                assert lines[1].strip() == 'alias = "envtechms"'.encode('utf-8')

    @mock.patch.dict(os.environ, {"test": "local", "FLASK_RUN_FROM_CLI": "true"})
    def test_deploy_remote_cmd(self, monkeypatch, example_dir, progressbar, capsys):
        info = CwsScriptInfo(project_dir='.')
        info.app_import_path = "command:app"
        app = info.load_app()
        with app.test_request_context() as app_ctx:
            options = {
                'project_dir': example_dir,
                'stage': 'dev',
                'timeout': 30,
                'memory_size': 100,
                'terraform_cloud': True,
                'terraform_organization': "CoWorks",
                'terraform_dir': ".",
                'terraform_refresh': False,
            }
            info = ScriptInfo(create_app=lambda: app)
            terraform_context = TerraformContext(info, CliCtxMokup())
            backend = TerraformBackend(terraform_context, None, **options)
            terraform = Terraform(backend, terraform_dir=PosixPath("terraform"), workspace="common")
            with tempfile.NamedTemporaryFile() as fp:
                terraform.generate_file("terraform.j2", fp.name, **options)
                fp.seek(0)
                lines = fp.readlines()
                assert len(lines) == 43
                print(lines)
                assert "TERRAFORM ON CLOUD" in lines[1].decode('utf-8')
                assert "CoWorks" in lines[6].decode('utf-8')

    @mock.patch.dict(os.environ, {"test": "local", "FLASK_RUN_FROM_CLI": "true"})
    def test_destroy_cmd(self, monkeypatch, example_dir, progressbar, capsys):
        monkeypatch.setattr(boto3, "Session", mock.Mock(return_value=mock.Mock(return_value='region')))
        info = CwsScriptInfo(project_dir='.')
        info.app_import_path = "command:app"
        app = info.load_app()
        with app.test_request_context() as app_ctx:
            options = {
                'project_dir': example_dir,
                'stage': 'dev',
                'timeout': 30,
                'memory_size': 100,
                'terraform_dir': ".",
                'terraform_refresh': False,
            }
            info = ScriptInfo(create_app=lambda: app)
            terraform_context = TerraformContext(info, CliCtxMokup())
            backend = TerraformBackend(terraform_context, None, **options)
            terraform = Terraform(backend, terraform_dir=PosixPath("terraform"), workspace="common")
            with tempfile.NamedTemporaryFile() as fp:
                terraform.generate_file("deploy.j2", fp.name, **options)
                fp.seek(0)
                lines = fp.readlines()
                assert len(lines) == 2049
                assert lines[1].strip() == 'alias = "envtechms"'.encode('utf-8')



import os
from unittest import mock
from unittest.mock import Mock

import click

from coworks.cws.client import client


class TestClass:

    def test_default_command(self):
        assert 'run' in client.commands
        assert 'shell' in client.commands
        assert 'routes' in client.commands

    @mock.patch.dict(os.environ, {"FLASK_APP": "complete:app"})
    def test_routes_command(self, monkeypatch, samples_docs_dir):
        mclick = Mock()
        monkeypatch.setattr(click, "echo", mclick)
        client.main(['routes'], 'cws', standalone_mode=False)
        mclick.assert_called()
        assert len(mclick.mock_calls) == 8
        out = [call.args[0].split(' ')[0] for call in mclick.mock_calls]
        assert 'Endpoint' in out
        assert 'admin.get_route' in out
        assert 'get' in out
        assert 'post' in out



from unittest.mock import Mock

from coworks import Blueprint
from coworks import entry


class BP(Blueprint):

    @entry
    def get_test(self, index):
        return f"blueprint BP {index}"

    @entry
    def get_extended_test(self, index):
        return f"blueprint extended test {index}"


class InitBP(BP):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.do_before_activation = Mock()
        self.do_after_activation = Mock()

        @self.before_app_request
        def before_activation():
            self.do_before_activation()

        @self.after_app_request
        def after_activation(response):
            self.do_after_activation(response)
            return response



from unittest import mock
from unittest.mock import MagicMock

from coworks import TechMicroService
from coworks import entry
from coworks.utils import BIZ_BUCKET_HEADER_KEY
from coworks.utils import BIZ_KEY_HEADER_KEY
from ..event import get_event


class AsyncMS(TechMicroService):

    @entry
    def get(self):
        return "ok"


class TestClass:
    def test_async_store(self, empty_aws_context):
        with mock.patch('boto3.session.Session') as session:
            session.side_effect = lambda: session
            session.client = MagicMock(side_effect=lambda _: session.client)
            app = AsyncMS()
            event = get_event('/', 'get')
            with app.cws_client(event, empty_aws_context) as c:
                event['headers']['InvocationType'.lower()] = 'Event'
                event['headers'][BIZ_BUCKET_HEADER_KEY.lower()] = 'bucket'
                event['headers'][BIZ_KEY_HEADER_KEY.lower()] = 'specific/key'
                app(event, empty_aws_context)
                session.assert_called_once()
                session.client.assert_called_once_with('s3')
                session.client.upload_fileobj.assert_called_once()



import pytest

from coworks import Blueprint
from coworks import TechMicroService
from coworks import entry


class BP(Blueprint):

    @entry
    def get_entry(self):
        return f"blueprint test entry"


class MS(TechMicroService):

    @entry
    def get_bp_entry(self):
        return f"blueprint test entry"


class TestClass:

    def test_overload(self):
        app = MS()
        app.register_blueprint(BP(), url_prefix='bp')
        with pytest.raises(AssertionError) as pytest_wrapped_e:
            with app.test_client() as c:
                response = c.get('/', headers={'Authorization': 'token'})
        assert pytest_wrapped_e.type == AssertionError
        assert pytest_wrapped_e.value.args[0] == 'Duplicate route /bp/entry'



from click import Command
from click import UsageError


def no_project_context(f):
    """Decorator to allow command without need to have a project dir defined."""
    setattr(f, '__need_project_context', False)
    return f


class CwsCommand(Command):

    def invoke(self, ctx):
        if getattr(self.callback, '__need_project_context', True) and not self._context_project_dir(ctx):
            raise UsageError(f"Project dir {self._context_project_dir(ctx)} not defined.")
        return super().invoke(ctx)

    @staticmethod
    def _context_project_dir(ctx):
        if ctx.parent is None:
            return ctx.params.get('project_dir')
        return CwsCommand._context_project_dir(ctx.parent)
